@@ -4,15 +4,16 @@
 use std::iter::FusedIterator;
 use std::slice::Iter;
 
+use ruff_python_ast::statement_visitor::{StatementVisitor, walk_stmt};
 use ruff_python_ast::{self as ast, Stmt, Suite};
 use ruff_python_parser::{Token, TokenKind, Tokens};
+use ruff_source_file::UniversalNewlineIterator;
 use ruff_text_size::{Ranged, TextSize};
 
-use ruff_python_ast::statement_visitor::{walk_stmt, StatementVisitor};
-use ruff_source_file::{Locator, UniversalNewlineIterator};
+use crate::Locator;
 
 /// Extract doc lines (standalone comments) from a token sequence.
-pub(crate) fn doc_lines_from_tokens(tokens: &Tokens) -> DocLines {
+pub(crate) fn doc_lines_from_tokens(tokens: &Tokens) -> DocLines<'_> {
     DocLines::new(tokens)
 }
 
@@ -24,7 +25,7 @@ pub(crate) struct DocLines<'a> {
 impl<'a> DocLines<'a> {
     fn new(tokens: &'a Tokens) -> Self {
         Self {
-            inner: tokens.up_to_first_unknown().iter(),
+            inner: tokens.iter(),
             prev: TextSize::default(),
         }
     }
@@ -72,6 +73,7 @@ impl StatementVisitor<'_> for StringLinesVisitor<'_> {
         if let Stmt::Expr(ast::StmtExpr {
             value: expr,
             range: _,
+            node_index: _,
         }) = stmt
         {
             if expr.is_string_literal_expr() {
