@@ -2,6 +2,7 @@ use std::fmt::{Debug, Formatter};
 use std::iter::FusedIterator;
 
 use ruff_python_ast::docstrings::{leading_space, leading_words};
+use ruff_python_semantic::Definition;
 use ruff_text_size::{Ranged, TextLen, TextRange, TextSize};
 use strum_macros::EnumIter;
 
@@ -130,39 +131,12 @@ impl SectionKind {
             Self::Yields => "Yields",
         }
     }
-
-    /// Returns `true` if a section can contain subsections, as in:
-    /// ```python
-    /// Yields
-    /// ------
-    /// int
-    ///     Description of the anonymous integer return value.
-    /// ```
-    ///
-    /// For NumPy, see: <https://numpydoc.readthedocs.io/en/latest/format.html>
-    ///
-    /// For Google, see: <https://google.github.io/styleguide/pyguide.html#38-comments-and-docstrings>
-    pub(crate) fn has_subsections(self) -> bool {
-        matches!(
-            self,
-            Self::Args
-                | Self::Arguments
-                | Self::OtherArgs
-                | Self::OtherParameters
-                | Self::OtherParams
-                | Self::Parameters
-                | Self::Raises
-                | Self::Returns
-                | Self::SeeAlso
-                | Self::Warns
-                | Self::Yields
-        )
-    }
 }
 
 pub(crate) struct SectionContexts<'a> {
     contexts: Vec<SectionContextData>,
     docstring: &'a Docstring<'a>,
+    style: SectionStyle,
 }
 
 impl<'a> SectionContexts<'a> {
@@ -194,6 +168,7 @@ impl<'a> SectionContexts<'a> {
                     last.as_ref(),
                     previous_line.as_ref(),
                     lines.peek(),
+                    docstring.definition,
                 ) {
                     if let Some(mut last) = last.take() {
                         last.range = TextRange::new(last.start(), line.start());
@@ -221,14 +196,19 @@ impl<'a> SectionContexts<'a> {
         Self {
             contexts,
             docstring,
+            style,
         }
     }
 
+    pub(crate) fn style(&self) -> SectionStyle {
+        self.style
+    }
+
     pub(crate) fn len(&self) -> usize {
         self.contexts.len()
     }
 
-    pub(crate) fn iter(&self) -> SectionContextsIter {
+    pub(crate) fn iter(&self) -> SectionContextsIter<'_> {
         SectionContextsIter {
             docstring_body: self.docstring.body(),
             inner: self.contexts.iter(),
@@ -273,7 +253,7 @@ impl<'a> Iterator for SectionContextsIter<'a> {
     }
 }
 
-impl<'a> DoubleEndedIterator for SectionContextsIter<'a> {
+impl DoubleEndedIterator for SectionContextsIter<'_> {
     fn next_back(&mut self) -> Option<Self::Item> {
         let back = self.inner.next_back()?;
         Some(SectionContext {
@@ -396,7 +376,7 @@ impl<'a> SectionContext<'a> {
         NewlineWithTrailingNewline::with_offset(lines, self.offset() + self.data.summary_full_end)
     }
 
-    fn following_lines_str(&self) -> &'a str {
+    pub(crate) fn following_lines_str(&self) -> &'a str {
         &self.docstring_body.as_str()[self.following_range_relative()]
     }
 
@@ -438,6 +418,7 @@ fn suspected_as_section(line: &str, style: SectionStyle) -> Option<SectionKind>
 }
 
 /// Check if the suspected context is really a section header.
+#[expect(clippy::too_many_arguments)]
 fn is_docstring_section(
     line: &Line,
     indent_size: TextSize,
@@ -446,7 +427,9 @@ fn is_docstring_section(
     previous_section: Option<&SectionContextData>,
     previous_line: Option<&Line>,
     next_line: Option<&Line>,
+    definition: &Definition<'_>,
 ) -> bool {
+    // for function definitions, track the known argument names for more accurate section detection.
     // Determine whether the current line looks like a section header, e.g., "Args:".
     let section_name_suffix = line[usize::from(indent_size + section_name_size)..].trim();
     let this_looks_like_a_section_name =
@@ -461,8 +444,7 @@ fn is_docstring_section(
         if next_line.is_empty() {
             false
         } else {
-            let next_line_is_underline = next_line.chars().all(|char| matches!(char, '-' | '='));
-            next_line_is_underline
+            next_line.chars().all(|char| matches!(char, '-' | '='))
         }
     });
     if next_line_is_underline {
@@ -470,7 +452,7 @@ fn is_docstring_section(
     }
 
     // Determine whether the previous line looks like the end of a paragraph.
-    let previous_line_looks_like_end_of_paragraph = previous_line.map_or(true, |previous_line| {
+    let previous_line_looks_like_end_of_paragraph = previous_line.is_none_or(|previous_line| {
         let previous_line = previous_line.trim();
         let previous_line_ends_with_punctuation = [',', ';', '.', '-', '\\', '/', ']', '}', ')']
             .into_iter()
@@ -503,44 +485,80 @@ fn is_docstring_section(
     // ```
     // However, if the header is an _exact_ match (like `Returns:`, as opposed to `returns:`), then
     // continue to treat it as a section header.
-    if section_kind.has_subsections() {
-        if let Some(previous_section) = previous_section {
-            let verbatim = &line[TextRange::at(indent_size, section_name_size)];
+    if let Some(previous_section) = previous_section {
+        let verbatim = &line[TextRange::at(indent_size, section_name_size)];
 
-            // If the section is more deeply indented, assume it's a subsection, as in:
-            // ```python
-            // def func(args: tuple[int]):
-            //     """Toggle the gizmo.
-            //
-            //     Args:
-            //         args: The arguments to the function.
-            //     """
-            // ```
-            if previous_section.indent_size < indent_size {
-                if section_kind.as_str() != verbatim {
-                    return false;
-                }
+        // If the section is more deeply indented, assume it's a subsection, as in:
+        // ```python
+        // def func(args: tuple[int]):
+        //     """Toggle the gizmo.
+        //
+        //     Args:
+        //         args: The arguments to the function.
+        //     """
+        // ```
+        // As noted above, an exact match for a section name (like the inner `Args:` below) is
+        // treated as a section header, unless the enclosing `Definition` is a function and contains
+        // a parameter with the same name, as in:
+        // ```python
+        // def func(Args: tuple[int]):
+        //     """Toggle the gizmo.
+        //
+        //     Args:
+        //         Args: The arguments to the function.
+        //     """
+        // ```
+        if previous_section.indent_size < indent_size {
+            let section_name = section_kind.as_str();
+            if section_name != verbatim || has_parameter(definition, section_name) {
+                return false;
             }
+        }
 
-            // If the section isn't underlined, and isn't title-cased, assume it's a subsection,
-            // as in:
-            // ```python
-            // def func(parameters: tuple[int]):
-            //     """Toggle the gizmo.
-            //
-            //     Parameters:
-            //     -----
-            //     parameters:
-            //         The arguments to the function.
-            //     """
-            // ```
-            if !next_line_is_underline && verbatim.chars().next().is_some_and(char::is_lowercase) {
-                if section_kind.as_str() != verbatim {
-                    return false;
-                }
+        // If the section has a preceding empty line, assume it's _not_ a subsection, as in:
+        // ```python
+        // def func(args: tuple[int]):
+        //     """Toggle the gizmo.
+        //
+        //     Args:
+        //         args: The arguments to the function.
+        //
+        //     returns:
+        //         The return value of the function.
+        //     """
+        // ```
+        if previous_line.is_some_and(|line| line.trim().is_empty()) {
+            return true;
+        }
+
+        // If the section isn't underlined, and isn't title-cased, assume it's a subsection,
+        // as in:
+        // ```python
+        // def func(parameters: tuple[int]):
+        //     """Toggle the gizmo.
+        //
+        //     Parameters:
+        //     -----
+        //     parameters:
+        //         The arguments to the function.
+        //     """
+        // ```
+        if !next_line_is_underline && verbatim.chars().next().is_some_and(char::is_lowercase) {
+            if section_kind.as_str() != verbatim {
+                return false;
             }
         }
     }
 
     true
 }
+
+/// Returns whether or not `definition` is a function definition and contains a parameter with the
+/// same name as `section_name`.
+fn has_parameter(definition: &Definition, section_name: &str) -> bool {
+    definition.as_function_def().is_some_and(|func| {
+        func.parameters
+            .iter()
+            .any(|param| param.name() == section_name)
+    })
+}
