@@ -1,6 +1,9 @@
 use ruff_macros::{ViolationMetadata, derive_message_formats};
+use ruff_python_ast as ast;
+use ruff_text_size::Ranged;
 
 use crate::Violation;
+use crate::checkers::ast::Checker;
 
 /// ## What it does
 /// Checks for `nonlocal` names without bindings.
@@ -31,6 +34,7 @@ use crate::Violation;
 /// - [Python documentation: The `nonlocal` statement](https://docs.python.org/3/reference/simple_stmts.html#nonlocal)
 /// - [PEP 3104 – Access to Names in Outer Scopes](https://peps.python.org/pep-3104/)
 #[derive(ViolationMetadata)]
+#[violation_metadata(stable_since = "v0.0.174")]
 pub(crate) struct NonlocalWithoutBinding {
     pub(crate) name: String,
 }
@@ -42,3 +46,19 @@ impl Violation for NonlocalWithoutBinding {
         format!("Nonlocal name `{name}` found without binding")
     }
 }
+
+/// PLE0117
+pub(crate) fn nonlocal_without_binding(checker: &Checker, nonlocal: &ast::StmtNonlocal) {
+    if !checker.semantic().scope_id.is_global() {
+        for name in &nonlocal.names {
+            if checker.semantic().nonlocal(name).is_none() {
+                checker.report_diagnostic(
+                    NonlocalWithoutBinding {
+                        name: name.to_string(),
+                    },
+                    name.range(),
+                );
+            }
+        }
+    }
+}
